"""This is a copy of sklearn/tests/test_public_functions.py. It can be
removed when we support scikit-learn >= 1.2.
"""
from importlib import import_module
from inspect import signature

import pytest

from imblearn.utils._param_validation import (
    generate_invalid_param_val,
    generate_valid_param,
    make_constraint,
)

PARAM_VALIDATION_FUNCTION_LIST = [
    "imblearn.datasets.fetch_datasets",
    "imblearn.datasets.make_imbalance",
    "imblearn.metrics.classification_report_imbalanced",
    "imblearn.metrics.geometric_mean_score",
    "imblearn.metrics.macro_averaged_mean_absolute_error",
    "imblearn.metrics.make_index_balanced_accuracy",
    "imblearn.metrics.sensitivity_specificity_support",
    "imblearn.metrics.sensitivity_score",
    "imblearn.metrics.specificity_score",
    "imblearn.pipeline.make_pipeline",
]


@pytest.mark.parametrize("func_module", PARAM_VALIDATION_FUNCTION_LIST)
def test_function_param_validation(func_module):
    """Check that an informative error is raised when the value of a parameter does not
    have an appropriate type or value.
    """
    module_name, func_name = func_module.rsplit(".", 1)
    module = import_module(module_name)
    func = getattr(module, func_name)

    func_sig = signature(func)
    func_params = [
        p.name
        for p in func_sig.parameters.values()
        if p.kind not in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
    ]
    parameter_constraints = getattr(func, "_skl_parameter_constraints")

    # Generate valid values for the required parameters
    # The parameters `*args` and `**kwargs` are ignored since we cannot generate
    # constraints.
    required_params = [
        p.name
        for p in func_sig.parameters.values()
        if p.default is p.empty and p.kind not in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
    ]
    valid_required_params = {}
    for param_name in required_params:
        if parameter_constraints[param_name] == "no_validation":
            valid_required_params[param_name] = 1
        else:
            valid_required_params[param_name] = generate_valid_param(
                make_constraint(parameter_constraints[param_name][0])
            )

    # check that there is a constraint for each parameter
    if func_params:
        validation_params = parameter_constraints.keys()
        unexpected_params = set(validation_params) - set(func_params)
        missing_params = set(func_params) - set(validation_params)
        err_msg = (
            "Mismatch between _parameter_constraints and the parameters of"
            f" {func_name}.\nConsider the unexpected parameters {unexpected_params} and"
            f" expected but missing parameters {missing_params}\n"
        )
        assert set(validation_params) == set(func_params), err_msg

    # this object does not have a valid type for sure for all params
    param_with_bad_type = type("BadType", (), {})()

    for param_name in func_params:
        constraints = parameter_constraints[param_name]

        if constraints == "no_validation":
            # This parameter is not validated
            continue

        match = (
            rf"The '{param_name}' parameter of {func_name} must be .* Got .* instead."
        )

        # First, check that the error is raised if param doesn't match any valid type.
        with pytest.raises(ValueError, match=match):
            func(**{**valid_required_params, param_name: param_with_bad_type})

        # Then, for constraints that are more than a type constraint, check that the
        # error is raised if param does match a valid type but does not match any valid
        # value for this type.
        constraints = [make_constraint(constraint) for constraint in constraints]

        for constraint in constraints:
            try:
                bad_value = generate_invalid_param_val(constraint)
            except NotImplementedError:
                continue

            with pytest.raises(ValueError, match=match):
                func(**{**valid_required_params, param_name: bad_value})



"""Base class and original SMOTE methods for over-sampling"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Fernando Nogueira
#          Christos Aridas
#          Dzianis Dudnik
# License: MIT

import math
import numbers
import warnings

import numpy as np
import sklearn
from scipy import sparse
from sklearn.base import clone
from sklearn.exceptions import DataConversionWarning
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from sklearn.utils import (
    _safe_indexing,
    check_array,
    check_random_state,
)
from sklearn.utils.fixes import parse_version
from sklearn.utils.sparsefuncs_fast import (
    csr_mean_variance_axis0,
)
from sklearn.utils.validation import _num_features

from ...metrics.pairwise import ValueDifferenceMetric
from ...utils import Substitution, check_neighbors_object, check_target_type
from ...utils._docstring import _n_jobs_docstring, _random_state_docstring
from ...utils._param_validation import HasMethods, Interval, StrOptions
from ...utils._validation import _check_X
from ...utils.fixes import _is_pandas_df, _mode
from ..base import BaseOverSampler

sklearn_version = parse_version(sklearn.__version__).base_version
if parse_version(sklearn_version) < parse_version("1.5"):
    from sklearn.utils import _get_column_indices
else:
    from sklearn.utils._indexing import _get_column_indices


class BaseSMOTE(BaseOverSampler):
    """Base class for the different SMOTE algorithms."""

    _parameter_constraints: dict = {
        **BaseOverSampler._parameter_constraints,
        "k_neighbors": [
            Interval(numbers.Integral, 1, None, closed="left"),
            HasMethods(["kneighbors", "kneighbors_graph"]),
        ],
        "n_jobs": [numbers.Integral, None],
    }

    def __init__(
        self,
        sampling_strategy="auto",
        random_state=None,
        k_neighbors=5,
        n_jobs=None,
    ):
        super().__init__(sampling_strategy=sampling_strategy)
        self.random_state = random_state
        self.k_neighbors = k_neighbors
        self.n_jobs = n_jobs

    def _validate_estimator(self):
        """Check the NN estimators shared across the different SMOTE
        algorithms.
        """
        self.nn_k_ = check_neighbors_object(
            "k_neighbors", self.k_neighbors, additional_neighbor=1
        )

    def _make_samples(
        self, X, y_dtype, y_type, nn_data, nn_num, n_samples, step_size=1.0, y=None
    ):
        """A support function that returns artificial samples constructed along
        the line connecting nearest neighbours.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Points from which the points will be created.

        y_dtype : dtype
            The data type of the targets.

        y_type : str or int
            The minority target value, just so the function can return the
            target values for the synthetic variables with correct length in
            a clear format.

        nn_data : ndarray of shape (n_samples_all, n_features)
            Data set carrying all the neighbours to be used

        nn_num : ndarray of shape (n_samples_all, k_nearest_neighbours)
            The nearest neighbours of each sample in `nn_data`.

        n_samples : int
            The number of samples to generate.

        step_size : float, default=1.0
            The step size to create samples.

        y : ndarray of shape (n_samples_all,), default=None
            The true target associated with `nn_data`. Used by Borderline SMOTE-2 to
            weight the distances in the sample generation process.

        Returns
        -------
        X_new : {ndarray, sparse matrix} of shape (n_samples_new, n_features)
            Synthetically generated samples.

        y_new : ndarray of shape (n_samples_new,)
            Target values for synthetic samples.
        """
        random_state = check_random_state(self.random_state)
        samples_indices = random_state.randint(low=0, high=nn_num.size, size=n_samples)

        # np.newaxis for backwards compatability with random_state
        steps = step_size * random_state.uniform(size=n_samples)[:, np.newaxis]
        rows = np.floor_divide(samples_indices, nn_num.shape[1])
        cols = np.mod(samples_indices, nn_num.shape[1])

        X_new = self._generate_samples(X, nn_data, nn_num, rows, cols, steps, y_type, y)
        y_new = np.full(n_samples, fill_value=y_type, dtype=y_dtype)
        return X_new, y_new

    def _generate_samples(
        self, X, nn_data, nn_num, rows, cols, steps, y_type=None, y=None
    ):
        r"""Generate a synthetic sample.

        The rule for the generation is:

        .. math::
           \mathbf{s_{s}} = \mathbf{s_{i}} + \mathcal{u}(0, 1) \times
           (\mathbf{s_{i}} - \mathbf{s_{nn}}) \,

        where \mathbf{s_{s}} is the new synthetic samples, \mathbf{s_{i}} is
        the current sample, \mathbf{s_{nn}} is a randomly selected neighbors of
        \mathbf{s_{i}} and \mathcal{u}(0, 1) is a random number between [0, 1).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Points from which the points will be created.

        nn_data : ndarray of shape (n_samples_all, n_features)
            Data set carrying all the neighbours to be used.

        nn_num : ndarray of shape (n_samples_all, k_nearest_neighbours)
            The nearest neighbours of each sample in `nn_data`.

        rows : ndarray of shape (n_samples,), dtype=int
            Indices pointing at feature vector in X which will be used
            as a base for creating new samples.

        cols : ndarray of shape (n_samples,), dtype=int
            Indices pointing at which nearest neighbor of base feature vector
            will be used when creating new samples.

        steps : ndarray of shape (n_samples,), dtype=float
            Step sizes for new samples.

        y_type : str, int or None, default=None
            Class label of the current target classes for which we want to generate
            samples.

        y : ndarray of shape (n_samples_all,), default=None
            The true target associated with `nn_data`. Used by Borderline SMOTE-2 to
            weight the distances in the sample generation process.

        Returns
        -------
        X_new : {ndarray, sparse matrix} of shape (n_samples, n_features)
            Synthetically generated samples.
        """
        diffs = nn_data[nn_num[rows, cols]] - X[rows]
        if y is not None:  # only entering for BorderlineSMOTE-2
            random_state = check_random_state(self.random_state)
            mask_pair_samples = y[nn_num[rows, cols]] != y_type
            diffs[mask_pair_samples] *= random_state.uniform(
                low=0.0, high=0.5, size=(mask_pair_samples.sum(), 1)
            )

        if sparse.issparse(X):
            sparse_func = type(X).__name__
            steps = getattr(sparse, sparse_func)(steps)
            X_new = X[rows] + steps.multiply(diffs)
        else:
            X_new = X[rows] + steps * diffs

        return X_new.astype(X.dtype)

    def _in_danger_noise(self, nn_estimator, samples, target_class, y, kind="danger"):
        """Estimate if a set of sample are in danger or noise.

        Used by BorderlineSMOTE and SVMSMOTE.

        Parameters
        ----------
        nn_estimator : estimator object
            An estimator that inherits from
            :class:`~sklearn.neighbors.base.KNeighborsMixin` use to determine
            if a sample is in danger/noise.

        samples : {array-like, sparse matrix} of shape (n_samples, n_features)
            The samples to check if either they are in danger or not.

        target_class : int or str
            The target corresponding class being over-sampled.

        y : array-like of shape (n_samples,)
            The true label in order to check the neighbour labels.

        kind : {'danger', 'noise'}, default='danger'
            The type of classification to use. Can be either:

            - If 'danger', check if samples are in danger,
            - If 'noise', check if samples are noise.

        Returns
        -------
        output : ndarray of shape (n_samples,)
            A boolean array where True refer to samples in danger or noise.
        """
        x = nn_estimator.kneighbors(samples, return_distance=False)[:, 1:]
        nn_label = (y[x] != target_class).astype(int)
        n_maj = np.sum(nn_label, axis=1)

        if kind == "danger":
            # Samples are in danger for m/2 <= m' < m
            return np.bitwise_and(
                n_maj >= (nn_estimator.n_neighbors - 1) / 2,
                n_maj < nn_estimator.n_neighbors - 1,
            )
        else:  # kind == "noise":
            # Samples are noise for m = m'
            return n_maj == nn_estimator.n_neighbors - 1


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    n_jobs=_n_jobs_docstring,
    random_state=_random_state_docstring,
)
class SMOTE(BaseSMOTE):
    """Class to perform over-sampling using SMOTE.

    This object is an implementation of SMOTE - Synthetic Minority
    Over-sampling Technique as presented in [1]_.

    Read more in the :ref:`User Guide <smote_adasyn>`.

    Parameters
    ----------
    {sampling_strategy}

    {random_state}

    k_neighbors : int or object, default=5
        The nearest neighbors used to define the neighborhood of samples to use
        to generate the synthetic samples. You can pass:

        - an `int` corresponding to the number of neighbors to use. A
          `~sklearn.neighbors.NearestNeighbors` instance will be fitted in this
          case.
        - an instance of a compatible nearest neighbors algorithm that should
          implement both methods `kneighbors` and `kneighbors_graph`. For
          instance, it could correspond to a
          :class:`~sklearn.neighbors.NearestNeighbors` but could be extended to
          any compatible class.

    {n_jobs}

        .. deprecated:: 0.10
           `n_jobs` has been deprecated in 0.10 and will be removed in 0.12.
           It was previously used to set `n_jobs` of nearest neighbors
           algorithm. From now on, you can pass an estimator where `n_jobs` is
           already set instead.

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.

    nn_k_ : estimator object
        Validated k-nearest neighbours created from the `k_neighbors` parameter.

    n_features_in_ : int
        Number of features in the input dataset.

        .. versionadded:: 0.9

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature
        names that are all strings.

        .. versionadded:: 0.10

    See Also
    --------
    SMOTENC : Over-sample using SMOTE for continuous and categorical features.

    SMOTEN : Over-sample using the SMOTE variant specifically for categorical
        features only.

    BorderlineSMOTE : Over-sample using the borderline-SMOTE variant.

    SVMSMOTE : Over-sample using the SVM-SMOTE variant.

    ADASYN : Over-sample using ADASYN.

    KMeansSMOTE : Over-sample applying a clustering before to oversample using
        SMOTE.

    Notes
    -----
    See the original papers: [1]_ for more details.

    Supports multi-class resampling. A one-vs.-rest scheme is used as
    originally proposed in [1]_.

    References
    ----------
    .. [1] N. V. Chawla, K. W. Bowyer, L. O.Hall, W. P. Kegelmeyer, "SMOTE:
       synthetic minority over-sampling technique," Journal of artificial
       intelligence research, 321-357, 2002.

    Examples
    --------
    >>> from collections import Counter
    >>> from sklearn.datasets import make_classification
    >>> from imblearn.over_sampling import SMOTE
    >>> X, y = make_classification(n_classes=2, class_sep=2,
    ... weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 900, 0: 100}})
    >>> sm = SMOTE(random_state=42)
    >>> X_res, y_res = sm.fit_resample(X, y)
    >>> print('Resampled dataset shape %s' % Counter(y_res))
    Resampled dataset shape Counter({{0: 900, 1: 900}})
    """

    def __init__(
        self,
        *,
        sampling_strategy="auto",
        random_state=None,
        k_neighbors=5,
        n_jobs=None,
    ):
        super().__init__(
            sampling_strategy=sampling_strategy,
            random_state=random_state,
            k_neighbors=k_neighbors,
            n_jobs=n_jobs,
        )

    def _fit_resample(self, X, y):
        # FIXME: to be removed in 0.12
        if self.n_jobs is not None:
            warnings.warn(
                "The parameter `n_jobs` has been deprecated in 0.10 and will be "
                "removed in 0.12. You can pass an nearest neighbors estimator where "
                "`n_jobs` is already set instead.",
                FutureWarning,
            )

        self._validate_estimator()

        X_resampled = [X.copy()]
        y_resampled = [y.copy()]

        for class_sample, n_samples in self.sampling_strategy_.items():
            if n_samples == 0:
                continue
            target_class_indices = np.flatnonzero(y == class_sample)
            X_class = _safe_indexing(X, target_class_indices)

            self.nn_k_.fit(X_class)
            nns = self.nn_k_.kneighbors(X_class, return_distance=False)[:, 1:]
            X_new, y_new = self._make_samples(
                X_class, y.dtype, class_sample, X_class, nns, n_samples, 1.0
            )
            X_resampled.append(X_new)
            y_resampled.append(y_new)

        if sparse.issparse(X):
            X_resampled = sparse.vstack(X_resampled, format=X.format)
        else:
            X_resampled = np.vstack(X_resampled)
        y_resampled = np.hstack(y_resampled)

        return X_resampled, y_resampled


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    n_jobs=_n_jobs_docstring,
    random_state=_random_state_docstring,
)
class SMOTENC(SMOTE):
    """Synthetic Minority Over-sampling Technique for Nominal and Continuous.

    Unlike :class:`SMOTE`, SMOTE-NC for dataset containing numerical and
    categorical features. However, it is not designed to work with only
    categorical features.

    Read more in the :ref:`User Guide <smote_adasyn>`.

    .. versionadded:: 0.4

    Parameters
    ----------
    categorical_features : "infer" or array-like of shape (n_cat_features,) or \
            (n_features,), dtype={{bool, int, str}}
        Specified which features are categorical. Can either be:

        - "auto" (default) to automatically detect categorical features. Only
          supported when `X` is a :class:`pandas.DataFrame` and it corresponds
          to columns that have a :class:`pandas.CategoricalDtype`;
        - array of `int` corresponding to the indices specifying the categorical
          features;
        - array of `str` corresponding to the feature names. `X` should be a pandas
          :class:`pandas.DataFrame` in this case.
        - mask array of shape (n_features, ) and ``bool`` dtype for which
          ``True`` indicates the categorical features.

    categorical_encoder : estimator, default=None
        One-hot encoder used to encode the categorical features. If `None`, a
        :class:`~sklearn.preprocessing.OneHotEncoder` is used with default parameters
        apart from `handle_unknown` which is set to 'ignore'.

    {sampling_strategy}

    {random_state}

    k_neighbors : int or object, default=5
        The nearest neighbors used to define the neighborhood of samples to use
        to generate the synthetic samples. You can pass:

        - an `int` corresponding to the number of neighbors to use. A
          `~sklearn.neighbors.NearestNeighbors` instance will be fitted in this
          case.
        - an instance of a compatible nearest neighbors algorithm that should
          implement both methods `kneighbors` and `kneighbors_graph`. For
          instance, it could correspond to a
          :class:`~sklearn.neighbors.NearestNeighbors` but could be extended to
          any compatible class.

    {n_jobs}

        .. deprecated:: 0.10
           `n_jobs` has been deprecated in 0.10 and will be removed in 0.12.
           It was previously used to set `n_jobs` of nearest neighbors
           algorithm. From now on, you can pass an estimator where `n_jobs` is
           already set instead.

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.

    nn_k_ : estimator object
        Validated k-nearest neighbours created from the `k_neighbors` parameter.

    ohe_ : :class:`~sklearn.preprocessing.OneHotEncoder`
        The one-hot encoder used to encode the categorical features.

        .. deprecated:: 0.11
           `ohe_` is deprecated in 0.11 and will be removed in 0.13. Use
           `categorical_encoder_` instead.

    categorical_encoder_ : estimator
        The encoder used to encode the categorical features.

    categorical_features_ : ndarray of shape (n_cat_features,), dtype=np.int64
        Indices of the categorical features.

    continuous_features_ : ndarray of shape (n_cont_features,), dtype=np.int64
        Indices of the continuous features.

    median_std_ : dict of int -> float
        Median of the standard deviation of the continuous features for each
        class to be over-sampled.

    n_features_ : int
        Number of features observed at `fit`.

    n_features_in_ : int
        Number of features in the input dataset.

        .. versionadded:: 0.9

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature
        names that are all strings.

        .. versionadded:: 0.10

    See Also
    --------
    SMOTE : Over-sample using SMOTE.

    SMOTEN : Over-sample using the SMOTE variant specifically for categorical
        features only.

    SVMSMOTE : Over-sample using SVM-SMOTE variant.

    BorderlineSMOTE : Over-sample using Borderline-SMOTE variant.

    ADASYN : Over-sample using ADASYN.

    KMeansSMOTE : Over-sample applying a clustering before to oversample using
        SMOTE.

    Notes
    -----
    See the original paper [1]_ for more details.

    Supports multi-class resampling. A one-vs.-rest scheme is used as
    originally proposed in [1]_.

    See
    :ref:`sphx_glr_auto_examples_over-sampling_plot_comparison_over_sampling.py`,
    and
    :ref:`sphx_glr_auto_examples_over-sampling_plot_illustration_generation_sample.py`.

    References
    ----------
    .. [1] N. V. Chawla, K. W. Bowyer, L. O.Hall, W. P. Kegelmeyer, "SMOTE:
       synthetic minority over-sampling technique," Journal of artificial
       intelligence research, 321-357, 2002.

    Examples
    --------
    >>> from collections import Counter
    >>> from numpy.random import RandomState
    >>> from sklearn.datasets import make_classification
    >>> from imblearn.over_sampling import SMOTENC
    >>> X, y = make_classification(n_classes=2, class_sep=2,
    ... weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    >>> print(f'Original dataset shape {{X.shape}}')
    Original dataset shape (1000, 20)
    >>> print(f'Original dataset samples per class {{Counter(y)}}')
    Original dataset samples per class Counter({{1: 900, 0: 100}})
    >>> # simulate the 2 last columns to be categorical features
    >>> X[:, -2:] = RandomState(10).randint(0, 4, size=(1000, 2))
    >>> sm = SMOTENC(random_state=42, categorical_features=[18, 19])
    >>> X_res, y_res = sm.fit_resample(X, y)
    >>> print(f'Resampled dataset samples per class {{Counter(y_res)}}')
    Resampled dataset samples per class Counter({{0: 900, 1: 900}})
    """

    _required_parameters = ["categorical_features"]

    _parameter_constraints: dict = {
        **SMOTE._parameter_constraints,
        "categorical_features": ["array-like", StrOptions({"auto"})],
        "categorical_encoder": [
            HasMethods(["fit_transform", "inverse_transform"]),
            None,
        ],
    }

    def __init__(
        self,
        categorical_features,
        *,
        categorical_encoder=None,
        sampling_strategy="auto",
        random_state=None,
        k_neighbors=5,
        n_jobs=None,
    ):
        super().__init__(
            sampling_strategy=sampling_strategy,
            random_state=random_state,
            k_neighbors=k_neighbors,
            n_jobs=n_jobs,
        )
        self.categorical_features = categorical_features
        self.categorical_encoder = categorical_encoder

    def _check_X_y(self, X, y):
        """Overwrite the checking to let pass some string for categorical
        features.
        """
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X = _check_X(X)
        self._check_n_features(X, reset=True)
        self._check_feature_names(X, reset=True)
        return X, y, binarize_y

    def _validate_column_types(self, X):
        """Compute the indices of the categorical and continuous features."""
        if self.categorical_features == "auto":
            if not _is_pandas_df(X):
                raise ValueError(
                    "When `categorical_features='auto'`, the input data "
                    f"should be a pandas.DataFrame. Got {type(X)} instead."
                )
            import pandas as pd  # safely import pandas now

            are_columns_categorical = np.array(
                [isinstance(col_dtype, pd.CategoricalDtype) for col_dtype in X.dtypes]
            )
            self.categorical_features_ = np.flatnonzero(are_columns_categorical)
            self.continuous_features_ = np.flatnonzero(~are_columns_categorical)
        else:
            self.categorical_features_ = np.array(
                _get_column_indices(X, self.categorical_features)
            )
            self.continuous_features_ = np.setdiff1d(
                np.arange(self.n_features_), self.categorical_features_
            )

    def _validate_estimator(self):
        super()._validate_estimator()
        if self.categorical_features_.size == self.n_features_in_:
            raise ValueError(
                "SMOTE-NC is not designed to work only with categorical "
                "features. It requires some numerical features."
            )
        elif self.categorical_features_.size == 0:
            raise ValueError(
                "SMOTE-NC is not designed to work only with numerical "
                "features. It requires some categorical features."
            )

    def _fit_resample(self, X, y):
        # FIXME: to be removed in 0.12
        if self.n_jobs is not None:
            warnings.warn(
                "The parameter `n_jobs` has been deprecated in 0.10 and will be "
                "removed in 0.12. You can pass an nearest neighbors estimator where "
                "`n_jobs` is already set instead.",
                FutureWarning,
            )

        self.n_features_ = _num_features(X)
        self._validate_column_types(X)
        self._validate_estimator()

        X_continuous = _safe_indexing(X, self.continuous_features_, axis=1)
        X_continuous = check_array(X_continuous, accept_sparse=["csr", "csc"])
        X_categorical = _safe_indexing(X, self.categorical_features_, axis=1)
        if X_continuous.dtype.name != "object":
            dtype_ohe = X_continuous.dtype
        else:
            dtype_ohe = np.float64

        if self.categorical_encoder is None:
            self.categorical_encoder_ = OneHotEncoder(
                handle_unknown="ignore", dtype=dtype_ohe
            )
        else:
            self.categorical_encoder_ = clone(self.categorical_encoder)

        # the input of the OneHotEncoder needs to be dense
        X_ohe = self.categorical_encoder_.fit_transform(
            X_categorical.toarray() if sparse.issparse(X_categorical) else X_categorical
        )
        if not sparse.issparse(X_ohe):
            X_ohe = sparse.csr_matrix(X_ohe, dtype=dtype_ohe)

        X_encoded = sparse.hstack((X_continuous, X_ohe), format="csr", dtype=dtype_ohe)
        X_resampled = [X_encoded.copy()]
        y_resampled = [y.copy()]

        # SMOTE resampling starts here
        self.median_std_ = {}
        for class_sample, n_samples in self.sampling_strategy_.items():
            if n_samples == 0:
                continue
            target_class_indices = np.flatnonzero(y == class_sample)
            X_class = _safe_indexing(X_encoded, target_class_indices)

            _, var = csr_mean_variance_axis0(
                X_class[:, : self.continuous_features_.size]
            )
            self.median_std_[class_sample] = np.median(np.sqrt(var))

            # In the edge case where the median of the std is equal to 0, the 1s
            # entries will be also nullified. In this case, we store the original
            # categorical encoding which will be later used for inverting the OHE
            if math.isclose(self.median_std_[class_sample], 0):
                # This variable will be used when generating data
                self._X_categorical_minority_encoded = X_class[
                    :, self.continuous_features_.size :
                ].toarray()

            # we can replace the 1 entries of the categorical features with the
            # median of the standard deviation. It will ensure that whenever
            # distance is computed between 2 samples, the difference will be equal
            # to the median of the standard deviation as in the original paper.
            X_class_categorical = X_class[:, self.continuous_features_.size :]
            # With one-hot encoding, the median will be repeated twice. We need
            # to divide by sqrt(2) such that we only have one median value
            # contributing to the Euclidean distance
            X_class_categorical.data[:] = self.median_std_[class_sample] / np.sqrt(2)
            X_class[:, self.continuous_features_.size :] = X_class_categorical

            self.nn_k_.fit(X_class)
            nns = self.nn_k_.kneighbors(X_class, return_distance=False)[:, 1:]
            X_new, y_new = self._make_samples(
                X_class, y.dtype, class_sample, X_class, nns, n_samples, 1.0
            )
            X_resampled.append(X_new)
            y_resampled.append(y_new)

        X_resampled = sparse.vstack(X_resampled, format=X_encoded.format)
        y_resampled = np.hstack(y_resampled)
        # SMOTE resampling ends here

        # reverse the encoding of the categorical features
        X_res_cat = X_resampled[:, self.continuous_features_.size :]
        X_res_cat.data = np.ones_like(X_res_cat.data)
        X_res_cat_dec = self.categorical_encoder_.inverse_transform(X_res_cat)

        if sparse.issparse(X):
            X_resampled = sparse.hstack(
                (
                    X_resampled[:, : self.continuous_features_.size],
                    X_res_cat_dec,
                ),
                format="csr",
            )
        else:
            X_resampled = np.hstack(
                (
                    X_resampled[:, : self.continuous_features_.size].toarray(),
                    X_res_cat_dec,
                )
            )

        indices_reordered = np.argsort(
            np.hstack((self.continuous_features_, self.categorical_features_))
        )
        if sparse.issparse(X_resampled):
            # the matrix is supposed to be in the CSR format after the stacking
            col_indices = X_resampled.indices.copy()
            for idx, col_idx in enumerate(indices_reordered):
                mask = X_resampled.indices == col_idx
                col_indices[mask] = idx
            X_resampled.indices = col_indices
        else:
            X_resampled = X_resampled[:, indices_reordered]

        return X_resampled, y_resampled

    def _generate_samples(self, X, nn_data, nn_num, rows, cols, steps, y_type, y=None):
        """Generate a synthetic sample with an additional steps for the
        categorical features.

        Each new sample is generated the same way than in SMOTE. However, the
        categorical features are mapped to the most frequent nearest neighbors
        of the majority class.
        """
        rng = check_random_state(self.random_state)
        X_new = super()._generate_samples(X, nn_data, nn_num, rows, cols, steps)
        # change in sparsity structure more efficient with LIL than CSR
        X_new = X_new.tolil() if sparse.issparse(X_new) else X_new

        # convert to dense array since scipy.sparse doesn't handle 3D
        nn_data = nn_data.toarray() if sparse.issparse(nn_data) else nn_data

        # In the case that the median std was equal to zeros, we have to
        # create non-null entry based on the encoded of OHE
        if math.isclose(self.median_std_[y_type], 0):
            nn_data[
                :, self.continuous_features_.size :
            ] = self._X_categorical_minority_encoded

        all_neighbors = nn_data[nn_num[rows]]

        categories_size = [self.continuous_features_.size] + [
            cat.size for cat in self.categorical_encoder_.categories_
        ]

        for start_idx, end_idx in zip(
            np.cumsum(categories_size)[:-1], np.cumsum(categories_size)[1:]
        ):
            col_maxs = all_neighbors[:, :, start_idx:end_idx].sum(axis=1)
            # tie breaking argmax
            is_max = np.isclose(col_maxs, col_maxs.max(axis=1, keepdims=True))
            max_idxs = rng.permutation(np.argwhere(is_max))
            xs, idx_sels = np.unique(max_idxs[:, 0], return_index=True)
            col_sels = max_idxs[idx_sels, 1]

            ys = start_idx + col_sels
            X_new[:, start_idx:end_idx] = 0
            X_new[xs, ys] = 1

        return X_new

    @property
    def ohe_(self):
        """One-hot encoder used to encode the categorical features."""
        warnings.warn(
            "'ohe_' attribute has been deprecated in 0.11 and will be removed "
            "in 0.13. Use 'categorical_encoder_' instead.",
            FutureWarning,
        )
        return self.categorical_encoder_


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    n_jobs=_n_jobs_docstring,
    random_state=_random_state_docstring,
)
class SMOTEN(SMOTE):
    """Synthetic Minority Over-sampling Technique for Nominal.

    This method is referred as SMOTEN in [1]_. It expects that the data to
    resample are only made of categorical features.

    Read more in the :ref:`User Guide <smote_adasyn>`.

    .. versionadded:: 0.8

    Parameters
    ----------
    categorical_encoder : estimator, default=None
        Ordinal encoder used to encode the categorical features. If `None`, a
        :class:`~sklearn.preprocessing.OrdinalEncoder` is used with default parameters.

    {sampling_strategy}

    {random_state}

    k_neighbors : int or object, default=5
        The nearest neighbors used to define the neighborhood of samples to use
        to generate the synthetic samples. You can pass:

        - an `int` corresponding to the number of neighbors to use. A
          `~sklearn.neighbors.NearestNeighbors` instance will be fitted in this
          case.
        - an instance of a compatible nearest neighbors algorithm that should
          implement both methods `kneighbors` and `kneighbors_graph`. For
          instance, it could correspond to a
          :class:`~sklearn.neighbors.NearestNeighbors` but could be extended to
          any compatible class.

    {n_jobs}

        .. deprecated:: 0.10
           `n_jobs` has been deprecated in 0.10 and will be removed in 0.12.
           It was previously used to set `n_jobs` of nearest neighbors
           algorithm. From now on, you can pass an estimator where `n_jobs` is
           already set instead.

    Attributes
    ----------
    categorical_encoder_ : estimator
        The encoder used to encode the categorical features.

    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.

    nn_k_ : estimator object
        Validated k-nearest neighbours created from the `k_neighbors` parameter.

    n_features_in_ : int
        Number of features in the input dataset.

        .. versionadded:: 0.9

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature
        names that are all strings.

        .. versionadded:: 0.10

    See Also
    --------
    SMOTE : Over-sample using SMOTE.

    SMOTENC : Over-sample using SMOTE for continuous and categorical features.

    BorderlineSMOTE : Over-sample using the borderline-SMOTE variant.

    SVMSMOTE : Over-sample using the SVM-SMOTE variant.

    ADASYN : Over-sample using ADASYN.

    KMeansSMOTE : Over-sample applying a clustering before to oversample using
        SMOTE.

    Notes
    -----
    See the original papers: [1]_ for more details.

    Supports multi-class resampling. A one-vs.-rest scheme is used as
    originally proposed in [1]_.

    References
    ----------
    .. [1] N. V. Chawla, K. W. Bowyer, L. O.Hall, W. P. Kegelmeyer, "SMOTE:
       synthetic minority over-sampling technique," Journal of artificial
       intelligence research, 321-357, 2002.

    Examples
    --------
    >>> import numpy as np
    >>> X = np.array(["A"] * 10 + ["B"] * 20 + ["C"] * 30, dtype=object).reshape(-1, 1)
    >>> y = np.array([0] * 20 + [1] * 40, dtype=np.int32)
    >>> from collections import Counter
    >>> print(f"Original class counts: {{Counter(y)}}")
    Original class counts: Counter({{1: 40, 0: 20}})
    >>> from imblearn.over_sampling import SMOTEN
    >>> sampler = SMOTEN(random_state=0)
    >>> X_res, y_res = sampler.fit_resample(X, y)
    >>> print(f"Class counts after resampling {{Counter(y_res)}}")
    Class counts after resampling Counter({{0: 40, 1: 40}})
    """

    _parameter_constraints: dict = {
        **SMOTE._parameter_constraints,
        "categorical_encoder": [
            HasMethods(["fit_transform", "inverse_transform"]),
            None,
        ],
    }

    def __init__(
        self,
        categorical_encoder=None,
        *,
        sampling_strategy="auto",
        random_state=None,
        k_neighbors=5,
        n_jobs=None,
    ):
        super().__init__(
            sampling_strategy=sampling_strategy,
            random_state=random_state,
            k_neighbors=k_neighbors,
            n_jobs=n_jobs,
        )
        self.categorical_encoder = categorical_encoder

    def _check_X_y(self, X, y):
        """Check should accept strings and not sparse matrices."""
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X, y = self._validate_data(
            X,
            y,
            reset=True,
            dtype=None,
            accept_sparse=["csr", "csc"],
        )
        return X, y, binarize_y

    def _validate_estimator(self):
        """Force to use precomputed distance matrix."""
        super()._validate_estimator()
        self.nn_k_.set_params(metric="precomputed")

    def _make_samples(self, X_class, klass, y_dtype, nn_indices, n_samples):
        random_state = check_random_state(self.random_state)
        # generate sample indices that will be used to generate new samples
        samples_indices = random_state.choice(
            np.arange(X_class.shape[0]), size=n_samples, replace=True
        )
        # for each drawn samples, select its k-neighbors and generate a sample
        # where for each feature individually, each category generated is the
        # most common category
        X_new = np.squeeze(
            _mode(X_class[nn_indices[samples_indices]], axis=1).mode, axis=1
        )
        y_new = np.full(n_samples, fill_value=klass, dtype=y_dtype)
        return X_new, y_new

    def _fit_resample(self, X, y):
        # FIXME: to be removed in 0.12
        if self.n_jobs is not None:
            warnings.warn(
                "The parameter `n_jobs` has been deprecated in 0.10 and will be "
                "removed in 0.12. You can pass an nearest neighbors estimator where "
                "`n_jobs` is already set instead.",
                FutureWarning,
            )

        if sparse.issparse(X):
            X_sparse_format = X.format
            X = X.toarray()
            warnings.warn(
                "Passing a sparse matrix to SMOTEN is not really efficient since it is"
                " converted to a dense array internally.",
                DataConversionWarning,
            )
        else:
            X_sparse_format = None

        self._validate_estimator()

        X_resampled = [X.copy()]
        y_resampled = [y.copy()]

        if self.categorical_encoder is None:
            self.categorical_encoder_ = OrdinalEncoder(dtype=np.int32)
        else:
            self.categorical_encoder_ = clone(self.categorical_encoder)
        X_encoded = self.categorical_encoder_.fit_transform(X)

        vdm = ValueDifferenceMetric(
            n_categories=[len(cat) for cat in self.categorical_encoder_.categories_]
        ).fit(X_encoded, y)

        for class_sample, n_samples in self.sampling_strategy_.items():
            if n_samples == 0:
                continue
            target_class_indices = np.flatnonzero(y == class_sample)
            X_class = _safe_indexing(X_encoded, target_class_indices)

            X_class_dist = vdm.pairwise(X_class)
            self.nn_k_.fit(X_class_dist)
            # the kneigbors search will include the sample itself which is
            # expected from the original algorithm
            nn_indices = self.nn_k_.kneighbors(X_class_dist, return_distance=False)
            X_new, y_new = self._make_samples(
                X_class, class_sample, y.dtype, nn_indices, n_samples
            )

            X_new = self.categorical_encoder_.inverse_transform(X_new)
            X_resampled.append(X_new)
            y_resampled.append(y_new)

        X_resampled = np.vstack(X_resampled)
        y_resampled = np.hstack(y_resampled)

        if X_sparse_format == "csr":
            return sparse.csr_matrix(X_resampled), y_resampled
        elif X_sparse_format == "csc":
            return sparse.csc_matrix(X_resampled), y_resampled
        else:
            return X_resampled, y_resampled

    def _more_tags(self):
        return {"X_types": ["2darray", "dataframe", "string"]}



"""Test the module one-sided selection."""
# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Christos Aridas
# License: MIT

import numpy as np
import pytest
from sklearn.datasets import make_classification
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils._testing import assert_array_equal

from imblearn.under_sampling import OneSidedSelection

RND_SEED = 0
X = np.array(
    [
        [-0.3879569, 0.6894251],
        [-0.09322739, 1.28177189],
        [-0.77740357, 0.74097941],
        [0.91542919, -0.65453327],
        [-0.03852113, 0.40910479],
        [-0.43877303, 1.07366684],
        [-0.85795321, 0.82980738],
        [-0.18430329, 0.52328473],
        [-0.30126957, -0.66268378],
        [-0.65571327, 0.42412021],
        [-0.28305528, 0.30284991],
        [0.20246714, -0.34727125],
        [1.06446472, -1.09279772],
        [0.30543283, -0.02589502],
        [-0.00717161, 0.00318087],
    ]
)
Y = np.array([0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0])


def test_oss_init():
    oss = OneSidedSelection(random_state=RND_SEED)

    assert oss.n_seeds_S == 1
    assert oss.n_jobs is None
    assert oss.random_state == RND_SEED


def test_oss_fit_resample():
    oss = OneSidedSelection(random_state=RND_SEED)
    X_resampled, y_resampled = oss.fit_resample(X, Y)

    X_gt = np.array(
        [
            [-0.3879569, 0.6894251],
            [0.91542919, -0.65453327],
            [-0.65571327, 0.42412021],
            [1.06446472, -1.09279772],
            [0.30543283, -0.02589502],
            [-0.00717161, 0.00318087],
            [-0.09322739, 1.28177189],
            [-0.77740357, 0.74097941],
            [-0.43877303, 1.07366684],
            [-0.85795321, 0.82980738],
            [-0.30126957, -0.66268378],
            [0.20246714, -0.34727125],
        ]
    )
    y_gt = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    assert_array_equal(X_resampled, X_gt)
    assert_array_equal(y_resampled, y_gt)


@pytest.mark.parametrize("n_neighbors", [1, KNeighborsClassifier(n_neighbors=1)])
def test_oss_with_object(n_neighbors):
    oss = OneSidedSelection(random_state=RND_SEED, n_neighbors=n_neighbors)
    X_resampled, y_resampled = oss.fit_resample(X, Y)

    X_gt = np.array(
        [
            [-0.3879569, 0.6894251],
            [0.91542919, -0.65453327],
            [-0.65571327, 0.42412021],
            [1.06446472, -1.09279772],
            [0.30543283, -0.02589502],
            [-0.00717161, 0.00318087],
            [-0.09322739, 1.28177189],
            [-0.77740357, 0.74097941],
            [-0.43877303, 1.07366684],
            [-0.85795321, 0.82980738],
            [-0.30126957, -0.66268378],
            [0.20246714, -0.34727125],
        ]
    )
    y_gt = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    assert_array_equal(X_resampled, X_gt)
    assert_array_equal(y_resampled, y_gt)
    knn = 1
    oss = OneSidedSelection(random_state=RND_SEED, n_neighbors=knn)
    X_resampled, y_resampled = oss.fit_resample(X, Y)
    assert_array_equal(X_resampled, X_gt)
    assert_array_equal(y_resampled, y_gt)


def test_one_sided_selection_multiclass():
    """Check the validity of the fitted attributes `estimators_`."""
    X, y = make_classification(
        n_samples=1_000,
        n_classes=4,
        weights=[0.1, 0.2, 0.2, 0.5],
        n_clusters_per_class=1,
        random_state=0,
    )
    oss = OneSidedSelection(random_state=RND_SEED)
    oss.fit_resample(X, y)

    assert len(oss.estimators_) == len(oss.sampling_strategy_)
    other_classes = []
    for est in oss.estimators_:
        assert est.classes_[0] == 0  # minority class
        assert est.classes_[1] in {1, 2, 3}  # other classes
        other_classes.append(est.classes_[1])
    assert len(set(other_classes)) == len(other_classes)


# TODO: remove in 0.14
def test_one_sided_selection_deprecation():
    """Check that we raise a FutureWarning when accessing the parameter `estimator_`."""
    oss = OneSidedSelection(random_state=RND_SEED)
    oss.fit_resample(X, Y)
    warn_msg = "`estimator_` attribute has been deprecated"
    with pytest.warns(FutureWarning, match=warn_msg):
        oss.estimator_



import numpy as np
import pytest
import sklearn
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
from sklearn.utils._testing import assert_array_equal
from sklearn.utils.fixes import parse_version

from imblearn.ensemble import RUSBoostClassifier

sklearn_version = parse_version(sklearn.__version__)


@pytest.fixture
def imbalanced_dataset():
    return make_classification(
        n_samples=10000,
        n_features=3,
        n_informative=2,
        n_redundant=0,
        n_repeated=0,
        n_classes=3,
        n_clusters_per_class=1,
        weights=[0.01, 0.05, 0.94],
        class_sep=0.8,
        random_state=0,
    )


@pytest.mark.parametrize("algorithm", ["SAMME", "SAMME.R"])
@pytest.mark.filterwarnings("ignore:The SAMME.R algorithm (the default) is")
def test_rusboost(imbalanced_dataset, algorithm):
    X, y = imbalanced_dataset
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, stratify=y, random_state=1
    )
    classes = np.unique(y)

    n_estimators = 500
    rusboost = RUSBoostClassifier(
        n_estimators=n_estimators, algorithm=algorithm, random_state=0
    )
    rusboost.fit(X_train, y_train)
    assert_array_equal(classes, rusboost.classes_)

    # check that we have an ensemble of samplers and estimators with a
    # consistent size
    assert len(rusboost.estimators_) > 1
    assert len(rusboost.estimators_) == len(rusboost.samplers_)
    assert len(rusboost.pipelines_) == len(rusboost.samplers_)

    # each sampler in the ensemble should have different random state
    assert len({sampler.random_state for sampler in rusboost.samplers_}) == len(
        rusboost.samplers_
    )
    # each estimator in the ensemble should have different random state
    assert len({est.random_state for est in rusboost.estimators_}) == len(
        rusboost.estimators_
    )

    # check the consistency of the feature importances
    assert len(rusboost.feature_importances_) == imbalanced_dataset[0].shape[1]

    # check the consistency of the prediction outpus
    y_pred = rusboost.predict_proba(X_test)
    assert y_pred.shape[1] == len(classes)
    assert rusboost.decision_function(X_test).shape[1] == len(classes)

    score = rusboost.score(X_test, y_test)
    assert score > 0.6, f"Failed with algorithm {algorithm} and score {score}"

    y_pred = rusboost.predict(X_test)
    assert y_pred.shape == y_test.shape


@pytest.mark.parametrize("algorithm", ["SAMME", "SAMME.R"])
@pytest.mark.filterwarnings("ignore:The SAMME.R algorithm (the default) is")
def test_rusboost_sample_weight(imbalanced_dataset, algorithm):
    X, y = imbalanced_dataset
    sample_weight = np.ones_like(y)
    rusboost = RUSBoostClassifier(algorithm=algorithm, random_state=0)

    # Predictions should be the same when sample_weight are all ones
    y_pred_sample_weight = rusboost.fit(X, y, sample_weight).predict(X)
    y_pred_no_sample_weight = rusboost.fit(X, y).predict(X)

    assert_array_equal(y_pred_sample_weight, y_pred_no_sample_weight)

    rng = np.random.RandomState(42)
    sample_weight = rng.rand(y.shape[0])
    y_pred_sample_weight = rusboost.fit(X, y, sample_weight).predict(X)

    with pytest.raises(AssertionError):
        assert_array_equal(y_pred_no_sample_weight, y_pred_sample_weight)



"""Implement generators for ``keras`` which will balance the data."""


# This is a trick to avoid an error during tests collection with pytest. We
# avoid the error when importing the package raise the error at the moment of
# creating the instance.
# This is a trick to avoid an error during tests collection with pytest. We
# avoid the error when importing the package raise the error at the moment of
# creating the instance.
def import_keras():
    """Try to import keras from keras and tensorflow.

    This is possible to import the sequence from keras or tensorflow.
    """

    def import_from_keras():
        try:
            import keras  # noqa

            if hasattr(keras.utils, "Sequence"):
                return (keras.utils.Sequence,), True
            else:
                return (keras.utils.data_utils.Sequence,), True
        except ImportError:
            return tuple(), False

    def import_from_tensforflow():
        try:
            from tensorflow import keras

            if hasattr(keras.utils, "Sequence"):
                return (keras.utils.Sequence,), True
            else:
                return (keras.utils.data_utils.Sequence,), True
        except ImportError:
            return tuple(), False

    ParentClassKeras, has_keras_k = import_from_keras()
    ParentClassTensorflow, has_keras_tf = import_from_tensforflow()
    has_keras = has_keras_k or has_keras_tf
    if has_keras:
        if has_keras_k:
            ParentClass = ParentClassKeras
        else:
            ParentClass = ParentClassTensorflow
    else:
        ParentClass = (object,)
    return ParentClass, has_keras


ParentClass, HAS_KERAS = import_keras()

from scipy.sparse import issparse  # noqa
from sklearn.base import clone  # noqa
from sklearn.utils import _safe_indexing  # noqa
from sklearn.utils import check_random_state  # noqa

from ..tensorflow import balanced_batch_generator as tf_bbg  # noqa
from ..under_sampling import RandomUnderSampler  # noqa
from ..utils import Substitution  # noqa
from ..utils._docstring import _random_state_docstring  # noqa


class BalancedBatchGenerator(*ParentClass):  # type: ignore
    """Create balanced batches when training a keras model.

    Create a keras ``Sequence`` which is given to ``fit``. The
    sampler defines the sampling strategy used to balance the dataset ahead of
    creating the batch. The sampler should have an attribute
    ``sample_indices_``.

    .. versionadded:: 0.4

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features)
        Original imbalanced dataset.

    y : ndarray of shape (n_samples,) or (n_samples, n_classes)
        Associated targets.

    sample_weight : ndarray of shape (n_samples,)
        Sample weight.

    sampler : sampler object, default=None
        A sampler instance which has an attribute ``sample_indices_``.
        By default, the sampler used is a
        :class:`~imblearn.under_sampling.RandomUnderSampler`.

    batch_size : int, default=32
        Number of samples per gradient update.

    keep_sparse : bool, default=False
        Either or not to conserve or not the sparsity of the input (i.e. ``X``,
        ``y``, ``sample_weight``). By default, the returned batches will be
        dense.

    random_state : int, RandomState instance or None, default=None
        Control the randomization of the algorithm:

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    sampler_ : sampler object
        The sampler used to balance the dataset.

    indices_ : ndarray of shape (n_samples, n_features)
        The indices of the samples selected during sampling.

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> iris = load_iris()
    >>> from imblearn.datasets import make_imbalance
    >>> class_dict = dict()
    >>> class_dict[0] = 30; class_dict[1] = 50; class_dict[2] = 40
    >>> X, y = make_imbalance(iris.data, iris.target, sampling_strategy=class_dict)
    >>> import tensorflow
    >>> y = tensorflow.keras.utils.to_categorical(y, 3)
    >>> model = tensorflow.keras.models.Sequential()
    >>> model.add(
    ...     tensorflow.keras.layers.Dense(
    ...         y.shape[1], input_dim=X.shape[1], activation='softmax'
    ...     )
    ... )
    >>> model.compile(optimizer='sgd', loss='categorical_crossentropy',
    ...               metrics=['accuracy'])
    >>> from imblearn.keras import BalancedBatchGenerator
    >>> from imblearn.under_sampling import NearMiss
    >>> training_generator = BalancedBatchGenerator(
    ...     X, y, sampler=NearMiss(), batch_size=10, random_state=42)
    >>> callback_history = model.fit(training_generator, epochs=10, verbose=0)
    """

    # flag for keras sequence duck-typing
    use_sequence_api = True

    def __init__(
        self,
        X,
        y,
        *,
        sample_weight=None,
        sampler=None,
        batch_size=32,
        keep_sparse=False,
        random_state=None,
    ):
        if not HAS_KERAS:
            raise ImportError("'No module named 'keras'")
        self.X = X
        self.y = y
        self.sample_weight = sample_weight
        self.sampler = sampler
        self.batch_size = batch_size
        self.keep_sparse = keep_sparse
        self.random_state = random_state
        self._sample()

    def _sample(self):
        random_state = check_random_state(self.random_state)
        if self.sampler is None:
            self.sampler_ = RandomUnderSampler(random_state=random_state)
        else:
            self.sampler_ = clone(self.sampler)
        self.sampler_.fit_resample(self.X, self.y)
        if not hasattr(self.sampler_, "sample_indices_"):
            raise ValueError("'sampler' needs to have an attribute 'sample_indices_'.")
        self.indices_ = self.sampler_.sample_indices_
        # shuffle the indices since the sampler are packing them by class
        random_state.shuffle(self.indices_)

    def __len__(self):
        return int(self.indices_.size // self.batch_size)

    def __getitem__(self, index):
        X_resampled = _safe_indexing(
            self.X,
            self.indices_[index * self.batch_size : (index + 1) * self.batch_size],
        )
        y_resampled = _safe_indexing(
            self.y,
            self.indices_[index * self.batch_size : (index + 1) * self.batch_size],
        )
        if issparse(X_resampled) and not self.keep_sparse:
            X_resampled = X_resampled.toarray()
        if self.sample_weight is not None:
            sample_weight_resampled = _safe_indexing(
                self.sample_weight,
                self.indices_[index * self.batch_size : (index + 1) * self.batch_size],
            )

        if self.sample_weight is None:
            return X_resampled, y_resampled
        else:
            return X_resampled, y_resampled, sample_weight_resampled


@Substitution(random_state=_random_state_docstring)
def balanced_batch_generator(
    X,
    y,
    *,
    sample_weight=None,
    sampler=None,
    batch_size=32,
    keep_sparse=False,
    random_state=None,
):
    """Create a balanced batch generator to train keras model.

    Returns a generator --- as well as the number of step per epoch --- which
    is given to ``fit``. The sampler defines the sampling strategy
    used to balance the dataset ahead of creating the batch. The sampler should
    have an attribute ``sample_indices_``.

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features)
        Original imbalanced dataset.

    y : ndarray of shape (n_samples,) or (n_samples, n_classes)
        Associated targets.

    sample_weight : ndarray of shape (n_samples,), default=None
        Sample weight.

    sampler : sampler object, default=None
        A sampler instance which has an attribute ``sample_indices_``.
        By default, the sampler used is a
        :class:`~imblearn.under_sampling.RandomUnderSampler`.

    batch_size : int, default=32
        Number of samples per gradient update.

    keep_sparse : bool, default=False
        Either or not to conserve or not the sparsity of the input (i.e. ``X``,
        ``y``, ``sample_weight``). By default, the returned batches will be
        dense.

    {random_state}

    Returns
    -------
    generator : generator of tuple
        Generate batch of data. The tuple generated are either (X_batch,
        y_batch) or (X_batch, y_batch, sampler_weight_batch).

    steps_per_epoch : int
        The number of samples per epoch. Required by ``fit_generator`` in
        keras.

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> X, y = load_iris(return_X_y=True)
    >>> from imblearn.datasets import make_imbalance
    >>> class_dict = dict()
    >>> class_dict[0] = 30; class_dict[1] = 50; class_dict[2] = 40
    >>> from imblearn.datasets import make_imbalance
    >>> X, y = make_imbalance(X, y, sampling_strategy=class_dict)
    >>> import tensorflow
    >>> y = tensorflow.keras.utils.to_categorical(y, 3)
    >>> model = tensorflow.keras.models.Sequential()
    >>> model.add(
    ...     tensorflow.keras.layers.Dense(
    ...         y.shape[1], input_dim=X.shape[1], activation='softmax'
    ...     )
    ... )
    >>> model.compile(optimizer='sgd', loss='categorical_crossentropy',
    ...               metrics=['accuracy'])
    >>> from imblearn.keras import balanced_batch_generator
    >>> from imblearn.under_sampling import NearMiss
    >>> training_generator, steps_per_epoch = balanced_batch_generator(
    ...     X, y, sampler=NearMiss(), batch_size=10, random_state=42)
    >>> callback_history = model.fit(training_generator,
    ...                              steps_per_epoch=steps_per_epoch,
    ...                              epochs=10, verbose=0)
    """

    return tf_bbg(
        X=X,
        y=y,
        sample_weight=sample_weight,
        sampler=sampler,
        batch_size=batch_size,
        keep_sparse=keep_sparse,
        random_state=random_state,
    )



"""Class to perform under-sampling based on the instance hardness
threshold."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Dayvid Oliveira
#          Christos Aridas
# License: MIT

import numbers
from collections import Counter

import numpy as np
from sklearn.base import clone, is_classifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble._base import _set_random_states
from sklearn.model_selection import StratifiedKFold, cross_val_predict
from sklearn.utils import _safe_indexing, check_random_state

from ...utils import Substitution
from ...utils._docstring import _n_jobs_docstring, _random_state_docstring
from ...utils._param_validation import HasMethods
from ..base import BaseUnderSampler


@Substitution(
    sampling_strategy=BaseUnderSampler._sampling_strategy_docstring,
    n_jobs=_n_jobs_docstring,
    random_state=_random_state_docstring,
)
class InstanceHardnessThreshold(BaseUnderSampler):
    """Undersample based on the instance hardness threshold.

    Read more in the :ref:`User Guide <instance_hardness_threshold>`.

    Parameters
    ----------
    estimator : estimator object, default=None
        Classifier to be used to estimate instance hardness of the samples.
        This classifier should implement `predict_proba`.

    {sampling_strategy}

    {random_state}

    cv : int, default=5
        Number of folds to be used when estimating samples' instance hardness.

    {n_jobs}

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        correspond to the class labels from which to sample and the values
        are the number of samples to sample.

    estimator_ : estimator object
        The validated classifier used to estimate the instance hardness of the samples.

    sample_indices_ : ndarray of shape (n_new_samples,)
        Indices of the samples selected.

        .. versionadded:: 0.4

    n_features_in_ : int
        Number of features in the input dataset.

        .. versionadded:: 0.9

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature
        names that are all strings.

        .. versionadded:: 0.10

    See Also
    --------
    NearMiss : Undersample based on near-miss search.

    RandomUnderSampler : Random under-sampling.

    Notes
    -----
    The method is based on [1]_.

    Supports multi-class resampling: from each class to be under-sampled, it
    retains the observations with the highest probability of being correctly
    classified.

    References
    ----------
    .. [1] D. Smith, Michael R., Tony Martinez, and Christophe Giraud-Carrier.
       "An instance level analysis of data complexity." Machine learning
       95.2 (2014): 225-256.

    Examples
    --------
    >>> from collections import Counter
    >>> from sklearn.datasets import make_classification
    >>> from imblearn.under_sampling import InstanceHardnessThreshold
    >>> X, y = make_classification(n_classes=2, class_sep=2,
    ... weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 900, 0: 100}})
    >>> iht = InstanceHardnessThreshold(random_state=42)
    >>> X_res, y_res = iht.fit_resample(X, y)
    >>> print('Resampled dataset shape %s' % Counter(y_res))
    Resampled dataset shape Counter({{1: 5..., 0: 100}})
    """

    _parameter_constraints: dict = {
        **BaseUnderSampler._parameter_constraints,
        "estimator": [
            HasMethods(["fit", "predict_proba"]),
            None,
        ],
        "cv": ["cv_object"],
        "n_jobs": [numbers.Integral, None],
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        *,
        estimator=None,
        sampling_strategy="auto",
        random_state=None,
        cv=5,
        n_jobs=None,
    ):
        super().__init__(sampling_strategy=sampling_strategy)
        self.random_state = random_state
        self.estimator = estimator
        self.cv = cv
        self.n_jobs = n_jobs

    def _validate_estimator(self, random_state):
        """Private function to create the classifier"""

        if (
            self.estimator is not None
            and is_classifier(self.estimator)
            and hasattr(self.estimator, "predict_proba")
        ):
            self.estimator_ = clone(self.estimator)
            _set_random_states(self.estimator_, random_state)

        elif self.estimator is None:
            self.estimator_ = RandomForestClassifier(
                n_estimators=100,
                random_state=self.random_state,
                n_jobs=self.n_jobs,
            )

    def _fit_resample(self, X, y):
        random_state = check_random_state(self.random_state)
        self._validate_estimator(random_state)

        target_stats = Counter(y)
        skf = StratifiedKFold(
            n_splits=self.cv,
            shuffle=True,
            random_state=random_state,
        )
        probabilities = cross_val_predict(
            self.estimator_,
            X,
            y,
            cv=skf,
            n_jobs=self.n_jobs,
            method="predict_proba",
        )
        probabilities = probabilities[range(len(y)), y]

        idx_under = np.empty((0,), dtype=int)

        for target_class in np.unique(y):
            if target_class in self.sampling_strategy_.keys():
                n_samples = self.sampling_strategy_[target_class]
                threshold = np.percentile(
                    probabilities[y == target_class],
                    (1.0 - (n_samples / target_stats[target_class])) * 100.0,
                )
                index_target_class = np.flatnonzero(
                    probabilities[y == target_class] >= threshold
                )
            else:
                index_target_class = slice(None)

            idx_under = np.concatenate(
                (
                    idx_under,
                    np.flatnonzero(y == target_class)[index_target_class],
                ),
                axis=0,
            )

        self.sample_indices_ = idx_under

        return _safe_indexing(X, idx_under), _safe_indexing(y, idx_under)

    def _more_tags(self):
        return {"sample_indices": True}



from collections import Counter

import pytest
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.utils._testing import assert_allclose, assert_array_equal

from imblearn.over_sampling import BorderlineSMOTE


@pytest.mark.parametrize("kind", ["borderline-1", "borderline-2"])
def test_borderline_smote_no_in_danger_samples(kind):
    """Check that the algorithm behave properly even on a dataset without any sample
    in danger.
    """
    X, y = make_classification(
        n_samples=500,
        n_features=2,
        n_informative=2,
        n_redundant=0,
        n_repeated=0,
        n_clusters_per_class=1,
        n_classes=3,
        weights=[0.1, 0.2, 0.7],
        class_sep=1.5,
        random_state=1,
    )
    smote = BorderlineSMOTE(kind=kind, m_neighbors=3, k_neighbors=5, random_state=0)
    X_res, y_res = smote.fit_resample(X, y)

    assert_allclose(X, X_res)
    assert_allclose(y, y_res)
    assert not smote.in_danger_indices


def test_borderline_smote_kind():
    """Check the behaviour of the `kind` parameter.

    In short, "borderline-2" generates sample closer to the boundary decision than
    "borderline-1". We generate an example where a logistic regression will perform
    worse on "borderline-2" than on "borderline-1".
    """
    X, y = make_classification(
        n_samples=500,
        n_features=2,
        n_informative=2,
        n_redundant=0,
        n_repeated=0,
        n_clusters_per_class=1,
        n_classes=3,
        weights=[0.1, 0.2, 0.7],
        class_sep=1.0,
        random_state=1,
    )
    smote = BorderlineSMOTE(
        kind="borderline-1", m_neighbors=9, k_neighbors=5, random_state=0
    )
    X_res_borderline_1, y_res_borderline_1 = smote.fit_resample(X, y)
    smote.set_params(kind="borderline-2")
    X_res_borderline_2, y_res_borderline_2 = smote.fit_resample(X, y)

    score_borderline_1 = (
        LogisticRegression()
        .fit(X_res_borderline_1, y_res_borderline_1)
        .score(X_res_borderline_1, y_res_borderline_1)
    )
    score_borderline_2 = (
        LogisticRegression()
        .fit(X_res_borderline_2, y_res_borderline_2)
        .score(X_res_borderline_2, y_res_borderline_2)
    )
    assert score_borderline_1 > score_borderline_2


def test_borderline_smote_in_danger():
    X, y = make_classification(
        n_samples=500,
        n_features=2,
        n_informative=2,
        n_redundant=0,
        n_repeated=0,
        n_clusters_per_class=1,
        n_classes=3,
        weights=[0.1, 0.2, 0.7],
        class_sep=0.8,
        random_state=1,
    )
    smote = BorderlineSMOTE(
        kind="borderline-1",
        m_neighbors=9,
        k_neighbors=5,
        random_state=0,
    )
    _, y_res_1 = smote.fit_resample(X, y)
    in_danger_indices_borderline_1 = smote.in_danger_indices
    smote.set_params(kind="borderline-2")
    _, y_res_2 = smote.fit_resample(X, y)
    in_danger_indices_borderline_2 = smote.in_danger_indices

    for key1, key2 in zip(
        in_danger_indices_borderline_1, in_danger_indices_borderline_2
    ):
        assert_array_equal(
            in_danger_indices_borderline_1[key1], in_danger_indices_borderline_2[key2]
        )
    assert len(in_danger_indices_borderline_1) == len(in_danger_indices_borderline_2)
    counter = Counter(y_res_1)
    assert counter[0] == counter[1] == counter[2]
    counter = Counter(y_res_2)
    assert counter[0] == counter[1] == counter[2]



"""This is a copy of sklearn/utils/_param_validation.py. It can be removed when
we support scikit-learn >= 1.2.
"""
# mypy: ignore-errors
import functools
import math
import operator
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from inspect import signature
from numbers import Integral, Real

import numpy as np
import sklearn
from scipy.sparse import csr_matrix, issparse
from sklearn.utils.fixes import parse_version

from .._config import config_context, get_config
from ..utils.fixes import _is_arraylike_not_scalar

sklearn_version = parse_version(sklearn.__version__)

if sklearn_version < parse_version("1.4"):

    class InvalidParameterError(ValueError, TypeError):
        """Custom exception to be raised when the parameter of a class/method/function
        does not have a valid type or value.
        """

        # Inherits from ValueError and TypeError to keep backward compatibility.

    def validate_parameter_constraints(parameter_constraints, params, caller_name):
        """Validate types and values of given parameters.

        Parameters
        ----------
        parameter_constraints : dict or {"no_validation"}
            If "no_validation", validation is skipped for this parameter.

            If a dict, it must be a dictionary `param_name: list of constraints`.
            A parameter is valid if it satisfies one of the constraints from the list.
            Constraints can be:
            - an Interval object, representing a continuous or discrete range of numbers
            - the string "array-like"
            - the string "sparse matrix"
            - the string "random_state"
            - callable
            - None, meaning that None is a valid value for the parameter
            - any type, meaning that any instance of this type is valid
            - an Options object, representing a set of elements of a given type
            - a StrOptions object, representing a set of strings
            - the string "boolean"
            - the string "verbose"
            - the string "cv_object"
            - the string "nan"
            - a MissingValues object representing markers for missing values
            - a HasMethods object, representing method(s) an object must have
            - a Hidden object, representing a constraint not meant to be exposed to the
              user

        params : dict
            A dictionary `param_name: param_value`. The parameters to validate against
            the constraints.

        caller_name : str
            The name of the estimator or function or method that called this function.
        """
        for param_name, param_val in params.items():
            # We allow parameters to not have a constraint so that third party
            # estimators can inherit from sklearn estimators without having to
            # necessarily use the validation tools.
            if param_name not in parameter_constraints:
                continue

            constraints = parameter_constraints[param_name]

            if constraints == "no_validation":
                continue

            constraints = [make_constraint(constraint) for constraint in constraints]

            for constraint in constraints:
                if constraint.is_satisfied_by(param_val):
                    # this constraint is satisfied, no need to check further.
                    break
            else:
                # No constraint is satisfied, raise with an informative message.

                # Ignore constraints that we don't want to expose in the error
                # message, i.e. options that are for internal purpose or not
                # officially supported.
                constraints = [
                    constraint for constraint in constraints if not constraint.hidden
                ]

                if len(constraints) == 1:
                    constraints_str = f"{constraints[0]}"
                else:
                    constraints_str = (
                        f"{', '.join([str(c) for c in constraints[:-1]])} or"
                        f" {constraints[-1]}"
                    )

                raise InvalidParameterError(
                    f"The {param_name!r} parameter of {caller_name} must be"
                    f" {constraints_str}. Got {param_val!r} instead."
                )

    def make_constraint(constraint):
        """Convert the constraint into the appropriate Constraint object.

        Parameters
        ----------
        constraint : object
            The constraint to convert.

        Returns
        -------
        constraint : instance of _Constraint
            The converted constraint.
        """
        if isinstance(constraint, str) and constraint == "array-like":
            return _ArrayLikes()
        if isinstance(constraint, str) and constraint == "sparse matrix":
            return _SparseMatrices()
        if isinstance(constraint, str) and constraint == "random_state":
            return _RandomStates()
        if constraint is callable:
            return _Callables()
        if constraint is None:
            return _NoneConstraint()
        if isinstance(constraint, type):
            return _InstancesOf(constraint)
        if isinstance(
            constraint, (Interval, StrOptions, Options, HasMethods, MissingValues)
        ):
            return constraint
        if isinstance(constraint, str) and constraint == "boolean":
            return _Booleans()
        if isinstance(constraint, str) and constraint == "verbose":
            return _VerboseHelper()
        if isinstance(constraint, str) and constraint == "cv_object":
            return _CVObjects()
        if isinstance(constraint, Hidden):
            constraint = make_constraint(constraint.constraint)
            constraint.hidden = True
            return constraint
        if isinstance(constraint, str) and constraint == "nan":
            return _NanConstraint()
        raise ValueError(f"Unknown constraint type: {constraint}")

    def validate_params(parameter_constraints, *, prefer_skip_nested_validation):
        """Decorator to validate types and values of functions and methods.

        Parameters
        ----------
        parameter_constraints : dict
            A dictionary `param_name: list of constraints`. See the docstring of
            `validate_parameter_constraints` for a description of the accepted
            constraints.

            Note that the *args and **kwargs parameters are not validated and must not
            be present in the parameter_constraints dictionary.

        prefer_skip_nested_validation : bool
            If True, the validation of parameters of inner estimators or functions
            called by the decorated function will be skipped.

            This is useful to avoid validating many times the parameters passed by the
            user from the public facing API. It's also useful to avoid validating
            parameters that we pass internally to inner functions that are guaranteed to
            be valid by the test suite.

            It should be set to True for most functions, except for those that receive
            non-validated objects as parameters or that are just wrappers around classes
            because they only perform a partial validation.

        Returns
        -------
        decorated_function : function or method
            The decorated function.
        """

        def decorator(func):
            # The dict of parameter constraints is set as an attribute of the function
            # to make it possible to dynamically introspect the constraints for
            # automatic testing.
            setattr(func, "_skl_parameter_constraints", parameter_constraints)

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                global_skip_validation = get_config()["skip_parameter_validation"]
                if global_skip_validation:
                    return func(*args, **kwargs)

                func_sig = signature(func)

                # Map *args/**kwargs to the function signature
                params = func_sig.bind(*args, **kwargs)
                params.apply_defaults()

                # ignore self/cls and positional/keyword markers
                to_ignore = [
                    p.name
                    for p in func_sig.parameters.values()
                    if p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
                ]
                to_ignore += ["self", "cls"]
                params = {
                    k: v for k, v in params.arguments.items() if k not in to_ignore
                }

                validate_parameter_constraints(
                    parameter_constraints, params, caller_name=func.__qualname__
                )

                try:
                    with config_context(
                        skip_parameter_validation=(
                            prefer_skip_nested_validation or global_skip_validation
                        )
                    ):
                        return func(*args, **kwargs)
                except InvalidParameterError as e:
                    # When the function is just a wrapper around an estimator, we allow
                    # the function to delegate validation to the estimator, but we
                    # replace the name of the estimator by the name of the function in
                    # the error message to avoid confusion.
                    msg = re.sub(
                        r"parameter of \w+ must be",
                        f"parameter of {func.__qualname__} must be",
                        str(e),
                    )
                    raise InvalidParameterError(msg) from e

            return wrapper

        return decorator

    class RealNotInt(Real):
        """A type that represents reals that are not instances of int.

        Behaves like float, but also works with values extracted from numpy arrays.
        isintance(1, RealNotInt) -> False
        isinstance(1.0, RealNotInt) -> True
        """

    RealNotInt.register(float)

    def _type_name(t):
        """Convert type into human readable string."""
        module = t.__module__
        qualname = t.__qualname__
        if module == "builtins":
            return qualname
        elif t == Real:
            return "float"
        elif t == Integral:
            return "int"
        return f"{module}.{qualname}"

    class _Constraint(ABC):
        """Base class for the constraint objects."""

        def __init__(self):
            self.hidden = False

        @abstractmethod
        def is_satisfied_by(self, val):
            """Whether or not a value satisfies the constraint.

            Parameters
            ----------
            val : object
                The value to check.

            Returns
            -------
            is_satisfied : bool
                Whether or not the constraint is satisfied by this value.
            """

        @abstractmethod
        def __str__(self):
            """A human readable representational string of the constraint."""

    class _InstancesOf(_Constraint):
        """Constraint representing instances of a given type.

        Parameters
        ----------
        type : type
            The valid type.
        """

        def __init__(self, type):
            super().__init__()
            self.type = type

        def is_satisfied_by(self, val):
            return isinstance(val, self.type)

        def __str__(self):
            return f"an instance of {_type_name(self.type)!r}"

    class _NoneConstraint(_Constraint):
        """Constraint representing the None singleton."""

        def is_satisfied_by(self, val):
            return val is None

        def __str__(self):
            return "None"

    class _NanConstraint(_Constraint):
        """Constraint representing the indicator `np.nan`."""

        def is_satisfied_by(self, val):
            return (
                not isinstance(val, Integral)
                and isinstance(val, Real)
                and math.isnan(val)
            )

        def __str__(self):
            return "numpy.nan"

    class _PandasNAConstraint(_Constraint):
        """Constraint representing the indicator `pd.NA`."""

        def is_satisfied_by(self, val):
            try:
                import pandas as pd

                return isinstance(val, type(pd.NA)) and pd.isna(val)
            except ImportError:
                return False

        def __str__(self):
            return "pandas.NA"

    class Options(_Constraint):
        """Constraint representing a finite set of instances of a given type.

        Parameters
        ----------
        type : type

        options : set
            The set of valid scalars.

        deprecated : set or None, default=None
            A subset of the `options` to mark as deprecated in the string
            representation of the constraint.
        """

        def __init__(self, type, options, *, deprecated=None):
            super().__init__()
            self.type = type
            self.options = options
            self.deprecated = deprecated or set()

            if self.deprecated - self.options:
                raise ValueError(
                    "The deprecated options must be a subset of the options."
                )

        def is_satisfied_by(self, val):
            return isinstance(val, self.type) and val in self.options

        def _mark_if_deprecated(self, option):
            """Add a deprecated mark to an option if needed."""
            option_str = f"{option!r}"
            if option in self.deprecated:
                option_str = f"{option_str} (deprecated)"
            return option_str

        def __str__(self):
            options_str = (
                f"{', '.join([self._mark_if_deprecated(o) for o in self.options])}"
            )
            return f"a {_type_name(self.type)} among {{{options_str}}}"

    class StrOptions(Options):
        """Constraint representing a finite set of strings.

        Parameters
        ----------
        options : set of str
            The set of valid strings.

        deprecated : set of str or None, default=None
            A subset of the `options` to mark as deprecated in the string
            representation of the constraint.
        """

        def __init__(self, options, *, deprecated=None):
            super().__init__(type=str, options=options, deprecated=deprecated)

    class Interval(_Constraint):
        """Constraint representing a typed interval.

        Parameters
        ----------
        type : {numbers.Integral, numbers.Real, RealNotInt}
            The set of numbers in which to set the interval.

            If RealNotInt, only reals that don't have the integer type
            are allowed. For example 1.0 is allowed but 1 is not.

        left : float or int or None
            The left bound of the interval. None means left bound is -∞.

        right : float, int or None
            The right bound of the interval. None means right bound is +∞.

        closed : {"left", "right", "both", "neither"}
            Whether the interval is open or closed. Possible choices are:

            - `"left"`: the interval is closed on the left and open on the right.
            It is equivalent to the interval `[ left, right )`.
            - `"right"`: the interval is closed on the right and open on the left.
            It is equivalent to the interval `( left, right ]`.
            - `"both"`: the interval is closed.
            It is equivalent to the interval `[ left, right ]`.
            - `"neither"`: the interval is open.
            It is equivalent to the interval `( left, right )`.

        Notes
        -----
        Setting a bound to `None` and setting the interval closed is valid. For
        instance, strictly speaking, `Interval(Real, 0, None, closed="both")`
        corresponds to `[0, +∞) U {+∞}`.
        """

        def __init__(self, type, left, right, *, closed):
            super().__init__()
            self.type = type
            self.left = left
            self.right = right
            self.closed = closed

            self._check_params()

        def _check_params(self):
            if self.type not in (Integral, Real, RealNotInt):
                raise ValueError(
                    "type must be either numbers.Integral, numbers.Real or RealNotInt."
                    f" Got {self.type} instead."
                )

            if self.closed not in ("left", "right", "both", "neither"):
                raise ValueError(
                    "closed must be either 'left', 'right', 'both' or 'neither'. "
                    f"Got {self.closed} instead."
                )

            if self.type is Integral:
                suffix = "for an interval over the integers."
                if self.left is not None and not isinstance(self.left, Integral):
                    raise TypeError(f"Expecting left to be an int {suffix}")
                if self.right is not None and not isinstance(self.right, Integral):
                    raise TypeError(f"Expecting right to be an int {suffix}")
                if self.left is None and self.closed in ("left", "both"):
                    raise ValueError(
                        f"left can't be None when closed == {self.closed} {suffix}"
                    )
                if self.right is None and self.closed in ("right", "both"):
                    raise ValueError(
                        f"right can't be None when closed == {self.closed} {suffix}"
                    )
            else:
                if self.left is not None and not isinstance(self.left, Real):
                    raise TypeError("Expecting left to be a real number.")
                if self.right is not None and not isinstance(self.right, Real):
                    raise TypeError("Expecting right to be a real number.")

            if (
                self.right is not None
                and self.left is not None
                and self.right <= self.left
            ):
                raise ValueError(
                    f"right can't be less than left. Got left={self.left} and "
                    f"right={self.right}"
                )

        def __contains__(self, val):
            if not isinstance(val, Integral) and np.isnan(val):
                return False

            left_cmp = operator.lt if self.closed in ("left", "both") else operator.le
            right_cmp = operator.gt if self.closed in ("right", "both") else operator.ge

            left = -np.inf if self.left is None else self.left
            right = np.inf if self.right is None else self.right

            if left_cmp(val, left):
                return False
            if right_cmp(val, right):
                return False
            return True

        def is_satisfied_by(self, val):
            if not isinstance(val, self.type):
                return False

            return val in self

        def __str__(self):
            type_str = "an int" if self.type is Integral else "a float"
            left_bracket = "[" if self.closed in ("left", "both") else "("
            left_bound = "-inf" if self.left is None else self.left
            right_bound = "inf" if self.right is None else self.right
            right_bracket = "]" if self.closed in ("right", "both") else ")"

            # better repr if the bounds were given as integers
            if not self.type == Integral and isinstance(self.left, Real):
                left_bound = float(left_bound)
            if not self.type == Integral and isinstance(self.right, Real):
                right_bound = float(right_bound)

            return (
                f"{type_str} in the range "
                f"{left_bracket}{left_bound}, {right_bound}{right_bracket}"
            )

    class _ArrayLikes(_Constraint):
        """Constraint representing array-likes"""

        def is_satisfied_by(self, val):
            return _is_arraylike_not_scalar(val)

        def __str__(self):
            return "an array-like"

    class _SparseMatrices(_Constraint):
        """Constraint representing sparse matrices."""

        def is_satisfied_by(self, val):
            return issparse(val)

        def __str__(self):
            return "a sparse matrix"

    class _Callables(_Constraint):
        """Constraint representing callables."""

        def is_satisfied_by(self, val):
            return callable(val)

        def __str__(self):
            return "a callable"

    class _RandomStates(_Constraint):
        """Constraint representing random states.

        Convenience class for
        [Interval(Integral, 0, 2**32 - 1, closed="both"), np.random.RandomState, None]
        """

        def __init__(self):
            super().__init__()
            self._constraints = [
                Interval(Integral, 0, 2**32 - 1, closed="both"),
                _InstancesOf(np.random.RandomState),
                _NoneConstraint(),
            ]

        def is_satisfied_by(self, val):
            return any(c.is_satisfied_by(val) for c in self._constraints)

        def __str__(self):
            return (
                f"{', '.join([str(c) for c in self._constraints[:-1]])} or"
                f" {self._constraints[-1]}"
            )

    class _Booleans(_Constraint):
        """Constraint representing boolean likes.

        Convenience class for
        [bool, np.bool_, Integral (deprecated)]
        """

        def __init__(self):
            super().__init__()
            self._constraints = [
                _InstancesOf(bool),
                _InstancesOf(np.bool_),
            ]

        def is_satisfied_by(self, val):
            return any(c.is_satisfied_by(val) for c in self._constraints)

        def __str__(self):
            return (
                f"{', '.join([str(c) for c in self._constraints[:-1]])} or"
                f" {self._constraints[-1]}"
            )

    class _VerboseHelper(_Constraint):
        """Helper constraint for the verbose parameter.

        Convenience class for
        [Interval(Integral, 0, None, closed="left"), bool, numpy.bool_]
        """

        def __init__(self):
            super().__init__()
            self._constraints = [
                Interval(Integral, 0, None, closed="left"),
                _InstancesOf(bool),
                _InstancesOf(np.bool_),
            ]

        def is_satisfied_by(self, val):
            return any(c.is_satisfied_by(val) for c in self._constraints)

        def __str__(self):
            return (
                f"{', '.join([str(c) for c in self._constraints[:-1]])} or"
                f" {self._constraints[-1]}"
            )

    class MissingValues(_Constraint):
        """Helper constraint for the `missing_values` parameters.

        Convenience for
        [
            Integral,
            Interval(Real, None, None, closed="both"),
            str,   # when numeric_only is False
            None,  # when numeric_only is False
            _NanConstraint(),
            _PandasNAConstraint(),
        ]

        Parameters
        ----------
        numeric_only : bool, default=False
            Whether to consider only numeric missing value markers.

        """

        def __init__(self, numeric_only=False):
            super().__init__()

            self.numeric_only = numeric_only

            self._constraints = [
                _InstancesOf(Integral),
                # we use an interval of Real to ignore np.nan that has its own
                # constraint
                Interval(Real, None, None, closed="both"),
                _NanConstraint(),
                _PandasNAConstraint(),
            ]
            if not self.numeric_only:
                self._constraints.extend([_InstancesOf(str), _NoneConstraint()])

        def is_satisfied_by(self, val):
            return any(c.is_satisfied_by(val) for c in self._constraints)

        def __str__(self):
            return (
                f"{', '.join([str(c) for c in self._constraints[:-1]])} or"
                f" {self._constraints[-1]}"
            )

    class HasMethods(_Constraint):
        """Constraint representing objects that expose specific methods.

        It is useful for parameters following a protocol and where we don't want to
        impose an affiliation to a specific module or class.

        Parameters
        ----------
        methods : str or list of str
            The method(s) that the object is expected to expose.
        """

        @validate_params(
            {"methods": [str, list]},
            prefer_skip_nested_validation=True,
        )
        def __init__(self, methods):
            super().__init__()
            if isinstance(methods, str):
                methods = [methods]
            self.methods = methods

        def is_satisfied_by(self, val):
            return all(callable(getattr(val, method, None)) for method in self.methods)

        def __str__(self):
            if len(self.methods) == 1:
                methods = f"{self.methods[0]!r}"
            else:
                methods = (
                    f"{', '.join([repr(m) for m in self.methods[:-1]])} and"
                    f" {self.methods[-1]!r}"
                )
            return f"an object implementing {methods}"

    class _IterablesNotString(_Constraint):
        """Constraint representing iterables that are not strings."""

        def is_satisfied_by(self, val):
            return isinstance(val, Iterable) and not isinstance(val, str)

        def __str__(self):
            return "an iterable"

    class _CVObjects(_Constraint):
        """Constraint representing cv objects.

        Convenient class for
        [
            Interval(Integral, 2, None, closed="left"),
            HasMethods(["split", "get_n_splits"]),
            _IterablesNotString(),
            None,
        ]
        """

        def __init__(self):
            super().__init__()
            self._constraints = [
                Interval(Integral, 2, None, closed="left"),
                HasMethods(["split", "get_n_splits"]),
                _IterablesNotString(),
                _NoneConstraint(),
            ]

        def is_satisfied_by(self, val):
            return any(c.is_satisfied_by(val) for c in self._constraints)

        def __str__(self):
            return (
                f"{', '.join([str(c) for c in self._constraints[:-1]])} or"
                f" {self._constraints[-1]}"
            )

    class Hidden:
        """Class encapsulating a constraint not meant to be exposed to the user.

        Parameters
        ----------
        constraint : str or _Constraint instance
            The constraint to be used internally.
        """

        def __init__(self, constraint):
            self.constraint = constraint

    def generate_invalid_param_val(constraint):
        """Return a value that does not satisfy the constraint.

        Raises a NotImplementedError if there exists no invalid value for this
        constraint.

        This is only useful for testing purpose.

        Parameters
        ----------
        constraint : _Constraint instance
            The constraint to generate a value for.

        Returns
        -------
        val : object
            A value that does not satisfy the constraint.
        """
        if isinstance(constraint, StrOptions):
            return f"not {' or '.join(constraint.options)}"

        if isinstance(constraint, MissingValues):
            return np.array([1, 2, 3])

        if isinstance(constraint, _VerboseHelper):
            return -1

        if isinstance(constraint, HasMethods):
            return type("HasNotMethods", (), {})()

        if isinstance(constraint, _IterablesNotString):
            return "a string"

        if isinstance(constraint, _CVObjects):
            return "not a cv object"

        if isinstance(constraint, Interval) and constraint.type is Integral:
            if constraint.left is not None:
                return constraint.left - 1
            if constraint.right is not None:
                return constraint.right + 1

            # There's no integer outside (-inf, +inf)
            raise NotImplementedError

        if isinstance(constraint, Interval) and constraint.type in (Real, RealNotInt):
            if constraint.left is not None:
                return constraint.left - 1e-6
            if constraint.right is not None:
                return constraint.right + 1e-6

            # bounds are -inf, +inf
            if constraint.closed in ("right", "neither"):
                return -np.inf
            if constraint.closed in ("left", "neither"):
                return np.inf

            # interval is [-inf, +inf]
            return np.nan

        raise NotImplementedError

    def generate_valid_param(constraint):
        """Return a value that does satisfy a constraint.

        This is only useful for testing purpose.

        Parameters
        ----------
        constraint : Constraint instance
            The constraint to generate a value for.

        Returns
        -------
        val : object
            A value that does satisfy the constraint.
        """
        if isinstance(constraint, _ArrayLikes):
            return np.array([1, 2, 3])

        if isinstance(constraint, _SparseMatrices):
            return csr_matrix([[0, 1], [1, 0]])

        if isinstance(constraint, _RandomStates):
            return np.random.RandomState(42)

        if isinstance(constraint, _Callables):
            return lambda x: x

        if isinstance(constraint, _NoneConstraint):
            return None

        if isinstance(constraint, _InstancesOf):
            if constraint.type is np.ndarray:
                # special case for ndarray since it can't be instantiated without
                # arguments
                return np.array([1, 2, 3])

            if constraint.type in (Integral, Real):
                # special case for Integral and Real since they are abstract classes
                return 1

            return constraint.type()

        if isinstance(constraint, _Booleans):
            return True

        if isinstance(constraint, _VerboseHelper):
            return 1

        if isinstance(constraint, MissingValues) and constraint.numeric_only:
            return np.nan

        if isinstance(constraint, MissingValues) and not constraint.numeric_only:
            return "missing"

        if isinstance(constraint, HasMethods):
            return type(
                "ValidHasMethods",
                (),
                {m: lambda self: None for m in constraint.methods},
            )()

        if isinstance(constraint, _IterablesNotString):
            return [1, 2, 3]

        if isinstance(constraint, _CVObjects):
            return 5

        if isinstance(constraint, Options):  # includes StrOptions
            for option in constraint.options:
                return option

        if isinstance(constraint, Interval):
            interval = constraint
            if interval.left is None and interval.right is None:
                return 0
            elif interval.left is None:
                return interval.right - 1
            elif interval.right is None:
                return interval.left + 1
            else:
                if interval.type is Real:
                    return (interval.left + interval.right) / 2
                else:
                    return interval.left + 1

        raise ValueError(f"Unknown constraint type: {constraint}")

else:
    from sklearn.utils._param_validation import generate_invalid_param_val  # noqa
    from sklearn.utils._param_validation import generate_valid_param  # noqa
    from sklearn.utils._param_validation import validate_parameter_constraints  # noqa
    from sklearn.utils._param_validation import (
        HasMethods,
        Hidden,
        Interval,
        InvalidParameterError,
        MissingValues,
        Options,
        RealNotInt,
        StrOptions,
        _ArrayLikes,
        _Booleans,
        _Callables,
        _CVObjects,
        _InstancesOf,
        _IterablesNotString,
        _NanConstraint,
        _NoneConstraint,
        _PandasNAConstraint,
        _RandomStates,
        _SparseMatrices,
        _VerboseHelper,
        make_constraint,
        validate_params,
    )
